package ru.dbotthepony.mc.otm.datagen

import net.minecraft.resources.ResourceLocation
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.IronBarsBlock
import net.minecraft.world.level.block.state.properties.BlockStateProperties
import net.minecraftforge.client.model.generators.ConfiguredModel
import net.minecraftforge.client.model.generators.ModelFile
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.fml.common.Mod
import net.minecraftforge.data.event.GatherDataEvent
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidResearchDataProvider
import ru.dbotthepony.mc.otm.block.*
import ru.dbotthepony.mc.otm.core.get
import ru.dbotthepony.mc.otm.datagen.blocks.BatteryBankProvider
import ru.dbotthepony.mc.otm.datagen.blocks.MatterBankProvider
import ru.dbotthepony.mc.otm.datagen.blocks.MatteryBlockStateProvider
import ru.dbotthepony.mc.otm.datagen.loot.LootModifiers
import ru.dbotthepony.mc.otm.datagen.items.MatteryItemModelProvider
import ru.dbotthepony.mc.otm.datagen.lang.AddEnglishLanguage
import ru.dbotthepony.mc.otm.datagen.loot.LootTables
import ru.dbotthepony.mc.otm.datagen.models.MatteryBlockModelProvider
import ru.dbotthepony.mc.otm.datagen.recipes.MatteryRecipeProvider
import ru.dbotthepony.mc.otm.registry.*
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.datagen.blocks.addBlockStates
import ru.dbotthepony.mc.otm.datagen.blocks.addComplexBlockStates
import ru.dbotthepony.mc.otm.datagen.items.addItemModels
import ru.dbotthepony.mc.otm.datagen.lang.MatteryLanguageProvider
import ru.dbotthepony.mc.otm.datagen.loot.addLootModifiers
import ru.dbotthepony.mc.otm.datagen.loot.addLootTables
import ru.dbotthepony.mc.otm.datagen.models.MatteryModelBuilder
import ru.dbotthepony.mc.otm.datagen.recipes.addBlastingRecipes
import ru.dbotthepony.mc.otm.datagen.recipes.addCraftingTableRecipes
import ru.dbotthepony.mc.otm.datagen.recipes.addDecorativesRecipes
import ru.dbotthepony.mc.otm.datagen.recipes.addPlatePressRecipes
import ru.dbotthepony.mc.otm.datagen.recipes.addShapelessRecipes
import ru.dbotthepony.mc.otm.datagen.tags.TagsProvider
import ru.dbotthepony.mc.otm.datagen.tags.addTags
import ru.dbotthepony.mc.otm.registry.objects.ColoredDecorativeBlock
import ru.dbotthepony.mc.otm.registry.objects.DecorativeBlock

fun modLocation(string: String) = ResourceLocation(DataGen.MOD_ID, string)

@Mod.EventBusSubscriber(modid = DataGen.MOD_ID, bus = Mod.EventBusSubscriber.Bus.MOD)
object DataGen {
	const val MOD_ID = OverdriveThatMatters.MOD_ID

	var blockModelProvider: MatteryBlockModelProvider by WriteOnce()
		private set
	var itemModelProvider: MatteryItemModelProvider by WriteOnce()
		private set
	var blockStateProvider: MatteryBlockStateProvider by WriteOnce()
		private set
	var lootTableProvider: LootTables by WriteOnce()
		private set
	var recipeProvider: MatteryRecipeProvider by WriteOnce()
		private set
	var lootModifier: LootModifiers by WriteOnce()
		private set
	var languageProvider: MatteryLanguageProvider by WriteOnce()
		private set

	fun decorativeCubeAll(vararg blocks: Block) {
		blockModelProvider.decorativeCubeAll(*blocks)
		blockStateProvider.simpleBlockM(*blocks)
	}

	fun decorativeCubeAll(subdir: String, vararg blocks: Block) {
		blockModelProvider.decorativeCubeAll(subdir, *blocks)
		blockStateProvider.simpleBlockM(*blocks)
	}

	fun decorativeCubeAll(subdir: String, suffix: String, vararg blocks: Block) {
		blockModelProvider.decorativeCubeAll(subdir, suffix, *blocks)
		blockStateProvider.simpleBlockM(*blocks)
	}

	fun decorativeCubeAll(blocks: List<Block>) {
		blockModelProvider.decorativeCubeAll(blocks)
		blockStateProvider.simpleBlockM(blocks)
	}

	fun resourceCubeAll(vararg blocks: Block) {
		blockModelProvider.resourceCubeAll(*blocks)
		blockStateProvider.simpleBlockM(*blocks)
	}

	fun decorativeColumn(block: Block, side: String, top: String) {
		blockModelProvider.decorativeColumn(block, side, top)
		blockStateProvider.simpleBlockM(block)
	}

	fun pane(block: Block, textureSide: ResourceLocation, textureRailing: ResourceLocation) {
		val name = block.registryName?.path ?: throw IllegalStateException("Invalid state of glass pane $block")

		val post = name
		val side = "${name}_ne"
		val side_alt = "${name}_sw"
		val noside = "${name}_noside_ne"
		val noside_alt = "${name}_noside_sw"

		var mdl_post: ModelFile? = null
		var mdl_side: ModelFile? = null
		var mdl_side_alt: ModelFile? = null
		var mdl_noside: ModelFile? = null
		var mdl_noside_alt: ModelFile? = null

		with(blockModelProvider) {
			exec {
				mdl_post = panePost(post, textureSide, textureRailing).renderType("translucent")
				mdl_side = paneSide(side, textureSide, textureRailing).renderType("translucent")
				mdl_side_alt = paneSideAlt(side_alt, textureSide, textureRailing).renderType("translucent")
				mdl_noside = paneNoSide(noside, textureSide).renderType("translucent")
				mdl_noside_alt = paneNoSideAlt(noside_alt, textureSide).renderType("translucent")
			}
		}

		blockStateProvider.exec {
			check(mdl_post != null) { "DataGen execution order is screwed lmaooooooo" }

			with(blockStateProvider.getMultipartBuilder(block)) {
				part().modelFile(mdl_post!!).addModel()
				part().modelFile(mdl_side!!).addModel().condition(IronBarsBlock.NORTH, true)
				part().modelFile(mdl_side!!).rotationY(90).addModel().condition(IronBarsBlock.EAST, true)
				part().modelFile(mdl_side_alt!!).addModel().condition(IronBarsBlock.SOUTH, true)
				part().modelFile(mdl_side_alt!!).rotationY(90).addModel().condition(IronBarsBlock.WEST, true)
				part().modelFile(mdl_noside!!).addModel().condition(IronBarsBlock.NORTH, false)
				part().modelFile(mdl_noside!!).rotationY(270).addModel().condition(IronBarsBlock.WEST, false)
				part().modelFile(mdl_noside_alt!!).addModel().condition(IronBarsBlock.EAST, false)
				part().modelFile(mdl_noside_alt!!).rotationY(90).addModel().condition(IronBarsBlock.SOUTH, false)
			}
		}
	}

	fun decoratives(list: ColoredDecorativeBlock) {
		for (block in list.blocks.values) {
			decorativeCubeAll(block)
		}

		for (item in list.items.values) {
			itemModelProvider.block(item)
		}
	}

	fun decoratives(list: DecorativeBlock) {
		for (block in list.allBlocks.values) {
			decorativeCubeAll(block)
		}

		for (item in list.allItems.values) {
			itemModelProvider.block(item)
		}
	}

	fun decoratives(subdir: String, list: ColoredDecorativeBlock) {
		for (block in list.blocks.values) {
			decorativeCubeAll(subdir, block)
		}

		for (item in list.items.values) {
			itemModelProvider.block(item)
		}
	}

	fun decoratives(subdir: String, list: DecorativeBlock) {
		for (block in list.allBlocks.values) {
			decorativeCubeAll(subdir, block)
		}

		for (item in list.allItems.values) {
			itemModelProvider.block(item)
		}
	}

	fun decoratives(subdir: String, suffix: String, list: ColoredDecorativeBlock) {
		for (block in list.blocks.values) {
			decorativeCubeAll(subdir, suffix, block)
		}

		for (item in list.items.values) {
			itemModelProvider.block(item)
		}
	}

	fun decoratives(subdir: String, suffix: String, list: DecorativeBlock) {
		for (block in list.allBlocks.values) {
			decorativeCubeAll(subdir, suffix, block)
		}

		for (item in list.allItems.values) {
			itemModelProvider.block(item)
		}
	}

	@SubscribeEvent
	@JvmStatic
	@Suppress("unused")
	fun onGatherData(event: GatherDataEvent) {
		val blockModelProvider = MatteryBlockModelProvider(event)
		val blockStateProvider = MatteryBlockStateProvider(event)
		val itemModelProvider = MatteryItemModelProvider(event)
		val lootTableProvider = LootTables(event.generator)
		val recipeProvider = MatteryRecipeProvider(event.generator)
		val lootModifier = LootModifiers(event.generator)
		val languageProvider = MatteryLanguageProvider(event.generator)

		this.blockModelProvider = blockModelProvider
		this.blockStateProvider = blockStateProvider
		this.itemModelProvider = itemModelProvider
		this.lootTableProvider = lootTableProvider
		this.recipeProvider = recipeProvider
		this.lootModifier = lootModifier
		this.languageProvider = languageProvider

		val tagsProvider = TagsProvider(event)

		addTags(tagsProvider)

		event.generator.addProvider(true, blockModelProvider)
		event.generator.addProvider(true, itemModelProvider)
		event.generator.addProvider(true, blockStateProvider)
		event.generator.addProvider(true, recipeProvider)
		event.generator.addProvider(true, MatterBankProvider(event))
		event.generator.addProvider(true, BatteryBankProvider(event))
		event.generator.addProvider(true, lootTableProvider)
		event.generator.addProvider(true, lootModifier)
		event.generator.addProvider(true, SoundDataProvider(event))
		event.generator.addProvider(true, AndroidResearchDataProvider(event.generator).also { it.exec { addResearchData(it, languageProvider) } })

		AddEnglishLanguage(languageProvider)

		blockModelProvider.resourceCubeAll(MBlocks.TRITANIUM_ORE)
		blockModelProvider.resourceCubeAll(MBlocks.TRITANIUM_RAW_BLOCK)
		blockModelProvider.resourceCubeAll(MBlocks.DEEPSLATE_TRITANIUM_ORE)

		with(blockModelProvider) {
			exec {
				copy("block/battery/battery1", "block/battery/battery0").also { it.offset(-4f, 0f, 0f) }
				copy("block/battery/battery2", "block/battery/battery0").also { it.offset(-8f, 0f, 0f) }

				copy("block/battery/battery3", "block/battery/battery0").also { it.offset(0f, 6f, 0f) }
				copy("block/battery/battery4", "block/battery/battery0").also { it.offset(-4f, 6f, 0f) }
				copy("block/battery/battery5", "block/battery/battery0").also { it.offset(-8f, 6f, 0f) }

				copy("block/battery/battery7", "block/battery/battery6").also { it.offset(4f, 0f, 0f) }
				copy("block/battery/battery8", "block/battery/battery6").also { it.offset(8f, 0f, 0f) }

				copy("block/battery/battery9", "block/battery/battery6").also { it.offset(0f, 6f, 0f) }
				copy("block/battery/battery10", "block/battery/battery6").also { it.offset(4f, 6f, 0f) }
				copy("block/battery/battery11", "block/battery/battery6").also { it.offset(8f, 6f, 0f) }

				for (i in 0 .. 11) {
					withExistingParent("block/battery/matter_capacitor$i", ResourceLocation(MOD_ID, "block/battery/battery$i"))
						.texture("1", "block/matterybank_core")
				}

				for (i in 1 .. 7) {
					copy("block/pattern/model$i", "block/pattern/model0").also { it.offset(-2f * i, 0f, 0f) }
				}
			}
		}

		addDecorativeData(blockStateProvider, itemModelProvider, blockModelProvider)

		addBlockStates(blockStateProvider)
		addItemModels(itemModelProvider)
		blockStateProvider.exec { addComplexBlockStates(blockStateProvider) }
		addLootTables(lootTableProvider)

		recipeProvider.exec { _, consumer ->
			addCraftingTableRecipes(consumer)
			addBlastingRecipes(consumer)
			addDecorativesRecipes(recipeProvider, consumer)
			addShapelessRecipes(consumer)
		}

		addPlatePressRecipes(recipeProvider)

		lootModifier.lambda {
			addLootModifiers(it)
		}

		languageProvider.registerProviders()
	}
}
