package ru.dbotthepony.mc.otm

import net.minecraftforge.common.ForgeConfigSpec
import kotlin.reflect.KProperty

operator fun <T> ForgeConfigSpec.ConfigValue<T>.setValue(config: Any, property: KProperty<*>, value: T) {
	set(value)
}

operator fun <T : Any> ForgeConfigSpec.ConfigValue<T>.getValue(config: Any, property: KProperty<*>): T {
	return get()
}

private val builderField by lazy {
	val field = ForgeConfigSpec.Builder::class.java.getDeclaredField("context")
	field.isAccessible = true
	field
}

private val builderContextField by lazy {
	val clazz = ForgeConfigSpec::class.java.declaredClasses.firstOrNull { it.name == "net.minecraftforge.common.ForgeConfigSpec\$BuilderContext" } ?: throw NoSuchElementException("Unable to find ForgeConfigSpec.BuilderContext")

	val field = clazz.getDeclaredField("comment")
	field.isAccessible = true
	field
}

fun ForgeConfigSpec.Builder.appendComment(vararg comments: String): ForgeConfigSpec.Builder {
	val context = builderField.get(this) ?: throw NullPointerException("No builder context!")
	val commentList = builderContextField.get(context) as? Array<String> ?: throw NullPointerException("BuilderContext.comment is null or not of right type!")

	val reconstruct = Array(commentList.size + comments.size) {
		if (it in commentList.indices) {
			commentList[it]
		} else {
			comments[it - commentList.size]
		}
	}

	builderContextField.set(context, reconstruct)
	return this
}

fun ForgeConfigSpec.Builder.defineInRange(path: String, value: Int, minValue: Int): ForgeConfigSpec.IntValue {
	return defineInRange(path, value, minValue, Int.MAX_VALUE)
}

fun ForgeConfigSpec.Builder.defineInRange(path: String, value: Double, minValue: Double): ForgeConfigSpec.DoubleValue {
	return defineInRange(path, value, minValue, Double.MAX_VALUE)
}
