package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.MenuProvider
import java.lang.Runnable
import net.minecraft.server.level.ServerLevel
import net.minecraft.client.multiplayer.ClientLevel
import net.minecraft.core.Direction
import net.minecraft.core.SectionPos
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.common.capabilities.ICapabilityProvider
import java.lang.ref.WeakReference
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.StringTag
import net.minecraft.network.chat.Component
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.ifHas
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.oncePre

abstract class MatteryBlockEntity(p_155228_: BlockEntityType<*>, p_155229_: BlockPos, p_155230_: BlockState) : BlockEntity(p_155228_, p_155229_, p_155230_), MenuProvider {
	var customDisplayName: Component? = null
	var redstoneSignal = 0
		set(level) {
			val old = isBlockedByRedstone
			field = level
			val state = isBlockedByRedstone

			if (old != state) {
				redstoneStatusUpdated(state, old)
			}
		}

	var redstoneSetting: RedstoneSetting = RedstoneSetting.LOW
		set(setting) {
			val old = isBlockedByRedstone
			field = setting
			val state = isBlockedByRedstone
			if (old != state) {
				redstoneStatusUpdated(state, old)
			}
		}

	val isBlockedByRedstone: Boolean
		get() {
			when (redstoneSetting) {
				RedstoneSetting.LOW -> {
					return redstoneSignal > 0
				}

				RedstoneSetting.HIGH -> {
					return redstoneSignal <= 0
				}

				RedstoneSetting.IGNORED -> {
					return false
				}
			}
		}

	protected open val defaultDisplayName: Component
		get() = level?.getBlockState(blockPos)?.block?.name ?: TextComponent("null at $blockPos")

	abstract override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu?

	protected fun tickOnce(func: Runnable) {
		level?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnceServer(func: Runnable) {
		(level as? ServerLevel)?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnceClient(func: Runnable) {
		(level as? ClientLevel)?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnce(func: (Level) -> Unit) {
		val level = level
		level?.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun tickOnceServer(func: (ServerLevel) -> Unit) {
		val level = level as? ServerLevel ?: return
		level.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun tickOnceClient(func: (ClientLevel) -> Unit) {
		val level = level as? ClientLevel ?: return
		level.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun <T> getAndBind(
		old: LazyOptional<T>,
		provider: ICapabilityProvider?,
		capability: Capability<T>,
		side: Direction,
		invalidate: Runnable
	): LazyOptional<T> {
		val get = provider?.getCapability(capability, side) ?: LazyOptional.empty()

		if (old !== get) {
			if (get.isPresent) {
				val ref = WeakReference(invalidate)
				get.addListener {
					ref.get()?.run()
				}
			}

			return get.cast()
		}

		return get.cast()
	}

	override fun getDisplayName(): Component {
		return customDisplayName ?: defaultDisplayName
	}

	protected open fun redstoneStatusUpdated(newBlocked: Boolean, oldBlocked: Boolean) {}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		val customDisplayName = customDisplayName
		if (customDisplayName != null)
			nbt.putString("Name", Component.Serializer.toJson(customDisplayName))

		nbt[REDSTONE_SETTING_KEY] = redstoneSetting.ordinal.toByte()
		nbt[REDSTONE_SIGNAL_KEY] = redstoneSignal.toByte()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("Name", StringTag::class.java) {
			customDisplayName = Component.Serializer.fromJson(it.asString)
		}

		redstoneSetting = RedstoneSetting[nbt.getByte(REDSTONE_SETTING_KEY).toInt()]
		redstoneSignal = nbt.getByte(REDSTONE_SIGNAL_KEY).toInt()
	}

	// Just to mark chunk unsaved
	open fun setChangedLight() {
		val level = level
		if (level is ServerLevel) {
			level.chunkSource.getChunkNow(
				SectionPos.blockToSectionCoord(blockPos.x),
				SectionPos.blockToSectionCoord(blockPos.z))?.isUnsaved = true
		}
	}

	companion object {
		const val REDSTONE_SETTING_KEY = "redstoneSetting"
		const val REDSTONE_SIGNAL_KEY = "redstoneSignal"
		const val INVENTORY_KEY = "container"
		const val MATTER_STORAGE_KEY = "matterStorage"
		const val ENERGY_KEY = "energyStorage"
		const val BATTERY_KEY = "batteryInventory"
	}
}
