package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.gui.screens.Screen
import net.minecraft.client.gui.screens.inventory.AbstractContainerScreen
import net.minecraft.client.gui.screens.inventory.InventoryScreen
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.player.Player
import ru.dbotthepony.mc.otm.client.render.element
import ru.dbotthepony.mc.otm.client.screen.ExoSuitInventoryScreen
import ru.dbotthepony.mc.otm.compat.cos.CosmeticToggleRenderButton
import ru.dbotthepony.mc.otm.compat.cos.isCosmeticArmorLoaded

private fun calculateScale(width: Float, height: Float): Int {
	val aspectRatio = width / height

	if (aspectRatio.isNaN()) {
		return 1
	}

	if (aspectRatio < 0.7083) {
		return (width / 51f * 30f).toInt()
	} else {
		return (height / 72f * 30f).toInt()
	}
}

class EntityRendererPanel<out S : Screen> @JvmOverloads constructor(
	screen: S,
	parent: EditablePanel<*>?,
	val entity: LivingEntity,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = ENTITY_RECTANGLE.width,
	height: Float = ENTITY_RECTANGLE.height,
	var renderScale: Int = calculateScale(width, height)
) : EditablePanel<S>(screen, parent, x, y, width, height) {
	init {
		dockPadding = DockProperty(1f, 1f, 1.5f, 1f)
	}

	val canvas = object : EditablePanel<S>(screen, this@EntityRendererPanel) {
		init {
			dock = Dock.FILL
			scissor = true
		}

		init {
			if (entity is Player && isCosmeticArmorLoaded) {
				CosmeticToggleRenderButton(screen, this, x = this.width - 7f, y = this.height - 7f)
			}
		}

		override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
			if (entity.isDeadOrDying) {
				return
			}

			val renderX = absoluteX.toInt() + width.toInt() / 2
			val renderY = absoluteY.toInt() + (height * 0.9f).toInt()

			InventoryScreen.renderEntityInInventory(
				renderX,
				renderY,
				renderScale,
				renderX - mouseX,
				absoluteY + height * 0.15f - mouseY,
				entity
			)
		}

		override fun performLayout() {
			super.performLayout()

			val button = children.firstOrNull { it is CosmeticToggleRenderButton } as CosmeticToggleRenderButton? ?: return

			button.x = this.width - 7f
			button.y = this.height - 7f
		}
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		ExoSuitInventoryScreen.ENTITY_RECTANGLE.render(stack)
	}

	companion object {
		val ENTITY_RECTANGLE = AbstractContainerScreen.INVENTORY_LOCATION.element(25f, 7f, 51f, 72f)
	}
}
