package ru.dbotthepony.mc.otm.data

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import com.google.gson.JsonSyntaxException
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.Serializer
import net.minecraft.world.level.storage.loot.parameters.LootContextParams
import net.minecraft.world.level.storage.loot.predicates.LootItemCondition
import net.minecraft.world.level.storage.loot.predicates.LootItemConditionType
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.capability.itemsStream
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.registry.MLootItemConditions

class ItemInInventoryCondition(
	val item: ItemStack,
	val matchDamage: Boolean = false,
	val matchNBT: Boolean = false,
	val matchCosmetics: Boolean = true,
) : LootItemCondition, LootItemCondition.Builder {
	override fun test(t: LootContext): Boolean {
		val matches = t[LootContextParams.LAST_DAMAGE_PLAYER]?.itemsStream(matchCosmetics)?.filter {
			if (it.isEmpty) {
				return@filter false
			}

			if (matchDamage && matchNBT) {
				it.item == item.item && it.tag == item.tag && it.damageValue == item.damageValue
			} else if (matchDamage) {
				it.item == item.item && it.damageValue == item.damageValue
			} else if (matchNBT) {
				it.item == item.item && it.tag == item.tag
			} else {
				it.item == item.item
			}
		} ?: return false

		var count = 0

		for (match in matches) {
			count += match.count

			if (count >= item.count) {
				return true
			}
		}

		return false
	}

	override fun getType(): LootItemConditionType {
		return MLootItemConditions.ITEM_IN_INVENTORY
	}

	override fun build(): LootItemCondition {
		return this
	}

	companion object : Serializer<ItemInInventoryCondition> {
		override fun serialize(
			p_79325_: JsonObject,
			p_79326_: ItemInInventoryCondition,
			p_79327_: JsonSerializationContext
		) {
			p_79325_["item"] = JsonPrimitive(p_79326_.item.item.registryName!!.toString())
			p_79325_["itemCount"] = JsonPrimitive(p_79326_.item.count)
			p_79325_["matchDamage"] = JsonPrimitive(p_79326_.matchDamage)
			p_79325_["matchNBT"] = JsonPrimitive(p_79326_.matchNBT)
			p_79325_["matchCosmetics"] = JsonPrimitive(p_79326_.matchCosmetics)
		}

		override fun deserialize(p_79323_: JsonObject, p_79324_: JsonDeserializationContext): ItemInInventoryCondition {
			val item = p_79323_["item"]?.asString ?: throw JsonSyntaxException("Missing item")
			val itemCount = p_79323_["itemCount"]?.asInt ?: throw JsonSyntaxException("Missing itemCount")
			val matchDamage = p_79323_["matchDamage"]?.asBoolean ?: false
			val matchNBT = p_79323_["matchNBT"]?.asBoolean ?: false
			val matchCosmetics = p_79323_["matchCosmetics"]?.asBoolean ?: true

			val getItem = ForgeRegistries.ITEMS.getValue(ResourceLocation(item))

			if (getItem == null || getItem == Items.AIR) {
				throw JsonSyntaxException("Invalid item $item")
			}

			return ItemInInventoryCondition(
				ItemStack(getItem, itemCount),
				matchDamage, matchNBT
			)
		}
	}
}
